/* Trainsched.c - display train schedule on Palm Pilot
 * (C) 2001 Philippe Marzouk <philm@users.sourceforge.net>
 */
/*
 This program is free software; you can redistribute it and/or modify
 it under the terms of the GNU General Public License as published by
 the Free Software Foundation; either version 2 of the License, or
 at your option) any later version.

 This program is distributed in the hope that it will be useful,
 but WITHOUT ANY WARRANTY; without even the implied warranty of
 MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 GNU General Public License for more details.

 You should have received a copy of the GNU General Public License
 along with this program; if not, write to the Free Software
 Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
#include <PalmOS.h>

#include "trainsched.h"

#define STATIONRECORD 0
#define DAYRECORD     1
#define OFFSETRECORD  2

/* To store program preferences */
typedef struct {
  UInt8 day;
  UInt8 from;
  UInt8 to;
  UInt16 position;
  char DBName[dmDBNameLength];
  LocalID pdbid;
  UInt16 CardNo;
} prefType;

/* to store Databases info */
typedef struct {
  char DBName[dmDBNameLength];
  LocalID pdbid;
  UInt16 CardNo;
} dbrowType;

/* Global variables */
LocalID pdbid=0;
UInt16 CardNo;
char DBName[dmDBNameLength]={0};
static MemHandle pdbarrayH=NULL;
static FormPtr gpForm = NULL;
static char *StationList[STATIONMAXNUM]={0};
static char StationNames[STATIONMAXNUM][STATIONMAXLENGTH+1]={{0},{0}};
static char *DayList[7]={0};
static char DayNames[7][DAYMAXLENGTH+1]={{0},{0}};
static UInt16 OffsetList[14]={0}; /* Max 7 days and 2 directions */
static UInt16 NumRecords=0;
static UInt16 numstations=0,numdays=0;
Int16 ScrollValue=0;
Int16 ScrollSelectValue=0;
UInt8 FromStation = NOSELECTION;
UInt8 ToStation = NOSELECTION;
UInt8 Day = NOSELECTION;
Int16 selpdbcount=0;

Boolean bUse3Cols = false;         /* Whether to use Annotation column. */
UInt16 nIDofTable = resTimeTable;  /* Which table to use 2 or 3 column one. */
Boolean bExternalAbbrevDb = false; /* Whether abbreviations from external db. */
#define MAX_ABBREVIATIONS 40
static char Abbreviations[MAX_ABBREVIATIONS][10]={{0},{0}}; /* Store abbreviations. */
static Char szAbbreviations[ 300 ]; 

/*
 * Function Prototypes
 */
static void TableDrawItemCallback(void * tableP, Int16 row, Int16 column,
                                   RectanglePtr bounds );
static Boolean MainFormDoCommand(UInt16 command);
void ApplicationDisplayDialog(UInt16 formID);
void SetFormLabel(FormPtr pForm);
void LoadSelTable(FormPtr pForm, Boolean Update);
void InitSelTable(FormPtr pForm);
void InitMainForm(FormPtr pForm);
void InitMainFormTable(FormPtr pForm);
static void LoadMainFormTable(FormPtr pForm);
Err WriteDBData(const void *dataP, UInt32 *sizeP, void *userDataP);

void ResetPrefs()
{
  FromStation = NOSELECTION;
  ToStation = NOSELECTION;
  Day = NOSELECTION;
  ScrollValue = 0;
  MemSet(DBName,dmDBNameLength,0);
  pdbid=0;
  CardNo=0;
}

static Err StartApplication() 
{
  prefType prefs;
  Int16 prefsVersion;
  UInt16 prefsSize;
  Boolean newSearch = true;
  DmSearchStateType searchState;
  EventType newEvent;
  MemHandle TextH;
  LocalID newID;
  UInt16 newCardNo;

  /* get previous user selection in the preferences */
  prefsSize = sizeof(prefType);
  prefsVersion=PrefGetAppPreferences(CREATORID,0,&prefs,&prefsSize,true);
  if(prefsVersion != noPreferenceFound) {
    FromStation = prefs.from;
    ToStation = prefs.to;
    Day = prefs.day;
    ScrollValue = prefs.position;
    CardNo = prefs.CardNo;
    MemMove (DBName,prefs.DBName,dmDBNameLength);

    /* First let's verify the PDB is still available */
    while(0 == DmGetNextDatabaseByTypeCreator(newSearch, &searchState,
                                               PDBTYPE, CREATORID,
                                               false, &newCardNo, &newID)) {
      newSearch = false;
      if (prefs.pdbid == 0)
        break;

      if (newID == prefs.pdbid) {
        pdbid = newID;
        break;
      }
    }

    /* saved database no more available */
    if (pdbid == 0)
      ResetPrefs();

    if (pdbid !=0)
      FrmGotoForm(resFormMain);

  }

  if (pdbid == 0) {
    newSearch = true;

    /* Open first database found else display an error */
    if (0 == DmGetNextDatabaseByTypeCreator(newSearch, &searchState,
                                             PDBTYPE, CREATORID,
                                             false, &CardNo, &pdbid)) {
      DmDatabaseInfo(CardNo,pdbid,DBName,NULL,NULL,NULL,NULL,NULL,
                      NULL,NULL,NULL,NULL,NULL);

      FrmGotoForm(resFormMain);
    } else {
      TextH = DmGetResource(strRsc, NO_PDB);
      FrmCustomAlert(alertDialog1, (char*) MemHandleLock(TextH), " ", " ");
      MemHandleUnlock(TextH);
      DmReleaseResource(TextH);

      ResetPrefs();

      /* Go back to Launcher */
      newEvent.eType = keyDownEvent;
      newEvent.data.keyDown.chr = launchChr;
      newEvent.data.keyDown.modifiers = commandKeyMask;
      EvtAddEventToQueue(&newEvent);
    }
  }
  return 0;
}

static void StopApplication() 
{
  prefType prefs;

  if (gpForm) {
    /* save user selection in the preferences */
    prefs.from = FromStation;
    prefs.to = ToStation;
    prefs.day = Day;
    prefs.position = ScrollValue;
    prefs.CardNo = CardNo;
    prefs.pdbid = pdbid;
    MemMove(prefs.DBName,DBName,dmDBNameLength);

    PrefSetAppPreferences(CREATORID,0,PREFAPPVERSION,&prefs,sizeof(prefType),true);
  }

  FrmCloseAllForms();
}

Boolean MainFormEventHandler(EventPtr event) 
{
  Boolean handled = false;
  Int16 val, min, max, page;

  switch (event->eType) {
  case frmOpenEvent: {
    gpForm = FrmGetActiveForm();

    InitMainForm(gpForm);

    InitMainFormTable(gpForm);

    LoadMainFormTable(gpForm);

    SetFormLabel(gpForm);

    handled = true;
    break;
  }

  case frmCloseEvent:
    if (gpForm) {
      FrmEraseForm(gpForm);
      FrmDeleteForm(gpForm);
      gpForm = 0;
    }
    handled = true;
    break;

  case frmUpdateEvent:
    FrmDrawForm(gpForm);
    InitMainForm(gpForm);
    LoadMainFormTable(gpForm);
    SetFormLabel(gpForm);
    handled=true;
    break;

  case winEnterEvent:
    handled = true;
    break;

  case popSelectEvent:
    switch(event->data.popSelect.listID) {
      case resFromList:
        if(FromStation != event->data.popSelect.selection) {
          FromStation  = event->data.popSelect.selection;
          LoadMainFormTable(gpForm);
        }
        break;

      case resToList:
        if (ToStation != event->data.popSelect.selection) {
          ToStation  = event->data.popSelect.selection;
          LoadMainFormTable(gpForm);
        }
        break;

      case resDayList:
        if(Day != event->data.popSelect.selection) {
          Day  = event->data.popSelect.selection;
          LoadMainFormTable(gpForm);
        }
        break;

      default:
        break;
    }

    break;
   
    /* Handle the scroll bar (should scroll the table) */
    /* FIXME need some work... in LoadMainFormTable */
    /*
  case sclRepeatEvent:
    if( ScrollValue != event->data.sclRepeat.newValue ) {
      ScrollValue = event->data.sclRepeat.newValue;
      LoadMainFormTable(gpForm);
    }
    break;
    */
  case sclExitEvent:
    if (ScrollValue != event->data.sclExit.newValue) {
      ScrollValue = event->data.sclExit.newValue;
      handled = 1;
      LoadMainFormTable(gpForm);
    }
    break;

    /* Make the physical up/down button scroll the list */
  case keyDownEvent:
    if (event->data.keyDown.modifiers & commandKeyMask) { /* physical key */
      if (event->data.keyDown.chr == chrPageUp) {
        gpForm = FrmGetActiveForm();
        SclGetScrollBar(FrmGetObjectPtr(gpForm, 
          FrmGetObjectIndex(gpForm, resTableScrollBar)),
                             &val, &min, &max, &page );
        if ( val >= min ) {
          if ((val - min) < 9) 
            ScrollValue -= val - min;
          else
            ScrollValue -= 9;
          LoadMainFormTable(gpForm);
        }
        handled = true;
      } else if (event->data.keyDown.chr == chrPageDown) {
        gpForm = FrmGetActiveForm();
        SclGetScrollBar(FrmGetObjectPtr(gpForm,
                         FrmGetObjectIndex(gpForm, resTableScrollBar)),
                                            &val, &min, &max, &page);
        if (val < max) {
          ScrollValue += 9;
          LoadMainFormTable(gpForm);
        }
        handled = true;
      }
    }
    break;
    
  case ctlSelectEvent: {
      Int16 temp = 0;
      EventType newevent;
      if (event->data.ctlSelect.controlID == resReverseBtn) {
        temp = ToStation;
        ToStation = FromStation;
        FromStation = temp;

        /* send a update event to redraw form*/
        MemSet(&newevent, sizeof(EventType), 0);
        newevent.eType = frmUpdateEvent;
//        newevent.data.frmUpdate.formID = gpForm->formId;
        newevent.data.frmUpdate.formID = FrmGetFormId(gpForm);
        newevent.data.frmUpdate.updateCode = frmRedrawUpdateCode;
        EvtAddEventToQueue(&newevent);
        
        handled=true;
      }
    }
    break;

  case nilEvent:
    handled = 1;
    break;

  default:
    /* -Wall warning eater: switch must handle all enum
     * elements. */
    break;
  }
  return handled;
}

Boolean aboutFormEventHandler(EventPtr event) 
{
  Boolean handled = false;
  
  FormPtr pForm = FrmInitForm(event->data.frmLoad.formID);
  
  switch (event->eType) {
    case frmOpenEvent:
      FrmDrawForm(pForm);
      handled=true;
      break;

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID) {
        case aboutFormOkButton: {
          /* send a close event */
          EventType event;
          MemSet(&event, sizeof(EventType), 0);
          event.eType = frmCloseEvent;
          event.data.frmClose.formID = event.data.frmLoad.formID;
          EvtAddEventToQueue(&event);
        }
        handled = true;
        break;

        default:
          break;
      }
    break;

    default:
      break;
  }
  return handled;
}

Boolean selectFormEventHandler(EventPtr event) 
{
  Boolean handled = false;
  FormPtr pForm;

  switch (event->eType) {
    /* Form Opening */
    case frmLoadEvent:
      pForm = FrmGetFormPtr (event->data.frmLoad.formID);
      FrmSetActiveForm(pForm);
      handled = true;
      break;

    case frmOpenEvent:
      pForm = FrmGetActiveForm();

      if (pdbarrayH) {
        MemHandleFree(pdbarrayH);
        pdbarrayH = NULL;
      }

      FrmDrawForm(pForm);
      InitSelTable(pForm);
      LoadSelTable(pForm, false);
      handled = true;
      break;

    case frmCloseEvent:
      pForm = FrmGetActiveForm ();

      if (pdbarrayH) {
        MemHandleFree(pdbarrayH);
        pdbarrayH = NULL;
      }

      FrmEraseForm(pForm);
      FrmDeleteForm(pForm);
      FrmGotoForm(resFormMain);
      handled = true;
      break;

    case winExitEvent:
      pForm = FrmGetActiveForm();

      if (pdbarrayH) {
        MemHandleFree(pdbarrayH);
        pdbarrayH = NULL;
      }

      FrmEraseForm(pForm);
      FrmDeleteForm(pForm);
      handled = true;
      break;

		case sclExitEvent:
			if (ScrollSelectValue != event->data.sclExit.newValue) {
				pForm = FrmGetActiveForm ();
				ScrollSelectValue = event->data.sclExit.newValue;
				handled = 1;
				MemHandleUnlock(pdbarrayH);
				LoadSelTable(pForm, true);
			}
			break;

    case ctlSelectEvent:
      switch (event->data.ctlSelect.controlID) {
        case selectFrmCancel: {
          /* send a close event */
          EventType event;

          MemSet(&event, sizeof(EventType), 0);
          event.eType = frmCloseEvent;
          event.data.frmClose.formID = event.data.frmLoad.formID;
          EvtAddEventToQueue(&event);

        }
          handled = true;
          break;

        case selectFrmOKButton: {
          /* send a close event */
          EventType event;
          Int16 rowP, columnP;
          dbrowType *data;
          TablePtr tableP;
          MemHandle TextH;

          pForm = FrmGetActiveForm ();

          tableP = FrmGetObjectPtr(pForm, 
                                    FrmGetObjectIndex(pForm,selectFrmTable));

          if (true == TblGetSelection(tableP, &rowP, &columnP)) {
            data = TblGetItemPtr (tableP, rowP, columnP);
            if (data) {
              ResetPrefs();
              pdbid = data->pdbid;
              CardNo = data->CardNo;
              MemMove (DBName,data->DBName,dmDBNameLength);
            }

            MemSet(&event, sizeof(EventType), 0);
            event.eType = frmCloseEvent;
            event.data.frmClose.formID = event.data.frmLoad.formID;
            EvtAddEventToQueue(&event);
          } else {
            TextH = DmGetResource(strRsc, NO_SELECTION);
            FrmCustomAlert(alertDialog1, (char*) MemHandleLock(TextH), " ", " ");
            MemHandleUnlock(TextH);
            DmReleaseResource(TextH);

            /* clear old events to avoid crashes with my 3.3 Rom 
             * maybe due to old events related to FrmCustomAlert 
             * not being cleared elsewhere, my fault ????? */
            do {
              EvtGetEvent(&event, 1);
              if (SysHandleEvent (&event)) continue;
            } while (event.eType != nilEvent);

            FrmEraseForm(pForm);
            FrmDeleteForm(pForm);
            FrmGotoForm(selectForm);
          }
        }
          handled = true;
          break;

        case selectFrmBeamButton: {
          Int16 rowP, columnP;
          dbrowType *data;
          TablePtr tableP;
          MemHandle TextH;
          EventType event;
          ExgSocketType exgSocket;
          Err err = 0;

          pForm = FrmGetActiveForm ();

          tableP = FrmGetObjectPtr(pForm, 
                                    FrmGetObjectIndex(pForm,selectFrmTable));

          if (true == TblGetSelection(tableP, &rowP, &columnP)) {

            data = TblGetItemPtr (tableP, rowP, columnP);
            if (data) {
              MemSet(&exgSocket, sizeof(exgSocket),0);
              exgSocket.description = data->DBName;
              exgSocket.name = "trainsched.pdb";

              err = ExgPut(&exgSocket);
              if (!err) {
                err = ExgDBWrite(WriteDBData, &exgSocket, data->DBName, data->pdbid, data->CardNo);
                err = ExgDisconnect(&exgSocket, err);
              }

              /* clear old events to avoid crashes with my 3.3 Rom 
               * there must be something wrong with the way I handle events 
               * or version 3.3 of PalmOS is very buggy... */
              do {
                EvtGetEvent(&event, 1);
                if (SysHandleEvent (&event)) continue;
              } while (event.eType != nilEvent);
            }
          } else {
            TextH = DmGetResource(strRsc, NO_SELECTION);
            FrmCustomAlert(alertDialog1, (char*) MemHandleLock(TextH), " ", " ");
            MemHandleUnlock(TextH);
            DmReleaseResource(TextH);

            /* clear old events to avoid crashes with my 3.3 Rom 
             * maybe due to old events related to FrmCustomAlert 
             * not being cleared elsewhere, my fault ????? */
            do {
              EvtGetEvent(&event, 1);
              if (SysHandleEvent (&event)) continue;
            } while (event.eType != nilEvent);

            FrmEraseForm(pForm);
            FrmDeleteForm(pForm);
            FrmGotoForm(selectForm);
          }
        }
          handled = true;
          break;

        case selectFrmDeleteButton: {
          Int16 rowP, columnP;
          dbrowType *data;
          TablePtr tableP;
          MemHandle TextH;
          EventType event;

          pForm = FrmGetActiveForm ();

          tableP = FrmGetObjectPtr(pForm, 
                                    FrmGetObjectIndex(pForm,selectFrmTable));
          if (true == TblGetSelection(tableP, &rowP, &columnP)) {
            data = TblGetItemPtr (tableP, rowP, columnP);
            if (data) {
              if ((pdbid == data->pdbid) && (CardNo == data->CardNo)) {

                TextH = DmGetResource(strRsc, NO_DELETE);
                FrmCustomAlert(alertDialog1, (char*) MemHandleLock(TextH), " ", " ");
                MemHandleUnlock(TextH);
                DmReleaseResource(TextH);

            /* clear old events to avoid crashes with my 3.3 Rom 
             * maybe due to old events related to FrmCustomAlert 
             * not being cleared elsewhere, my fault ????? */
            do {
              EvtGetEvent(&event, 1);
              if (SysHandleEvent (&event)) continue;
            } while (event.eType != nilEvent);

                FrmEraseForm(pForm);
                FrmDeleteForm(pForm);
                FrmGotoForm(selectForm);
              } else {
                /* FIXME error handling on delete */
                DmDeleteDatabase (data->CardNo, data->pdbid);

								LoadSelTable(pForm, false);
              }
            }
          } else {
            TextH = DmGetResource(strRsc, NO_SELECTION);
            FrmCustomAlert(alertDialog1, (char*) MemHandleLock(TextH), " ", " ");
            MemHandleUnlock(TextH);
            DmReleaseResource(TextH);

            /* clear old events to avoid crashes with my 3.3 Rom 
             * maybe due to old events related to FrmCustomAlert 
             * not being cleared elsewhere, my fault ????? */
            do {
              EvtGetEvent(&event, 1);
              if (SysHandleEvent (&event)) continue;
            } while (event.eType != nilEvent);

            FrmEraseForm(pForm);
            FrmDeleteForm(pForm);
            FrmGotoForm(selectForm);
          }
        }
          handled = true;
          break;

        default:
          break;
      }
    break;
/*
		case penDownEvent: {
			ControlType *CtlObject;
			
			pForm = FrmGetActiveForm();
			CtlObject = FrmGetObjectPtr (pForm, FrmGetObjectIndex (pForm, selectFrmOKButton));
			handled = CtlHandleEvent(CtlObject,event);
		  }
			break;
*/
    case nilEvent:
      handled = true;
      break;

    default:
      break;
  }
  return handled;
}

static Boolean ApplicationEventHandler(EventPtr event) 
{
  Boolean handled = false;
  
  switch (event->eType) {
    case frmLoadEvent: {
      FormPtr pForm = FrmInitForm(event->data.frmLoad.formID);

      switch (event->data.frmLoad.formID) {
        case resFormMain:
          FrmSetActiveForm(pForm);
          FrmSetEventHandler(pForm, (FormEventHandlerPtr)MainFormEventHandler);
          handled = true;
          break;

      case aboutForm:
          FrmSetEventHandler(pForm, (FormEventHandlerPtr)aboutFormEventHandler);
          handled = true;
          break;

        case selectForm:
          FrmSetActiveForm(pForm);
          FrmSetEventHandler(pForm, (FormEventHandlerPtr)selectFormEventHandler);
          handled = true;

        default:
          break;
      }
    }
    break;

    case menuOpenEvent:
      if ( bUse3Cols )
          MenuShowItem( MainOptionsAbbreviations );
      else
          MenuHideItem( MainOptionsAbbreviations );
      break;

    case menuEvent:
      return MainFormDoCommand (event->data.menu.itemID);
      break;

    case winEnterEvent: {
      if (event->data.winEnter.enterWindow == (WinHandle) FrmGetFormPtr(selectForm))
        handled = true;
    }
    break;

    default:
      /* -Wall warning eater: switch must handle all enum
       * elements. */
      break;
  }
  return handled;
}

static void EventLoop() 
{
  EventType event;
  UInt16 error;
  do {
    EvtGetEvent(&event, evtWaitForever);

    if (!SysHandleEvent(&event)) {
      if (!MenuHandleEvent(0, &event, &error)) {
        if (!ApplicationEventHandler(&event)) {
          FrmDispatchEvent(&event);
        }
      }
    }
  }
  while (event.eType != appStopEvent);
}

UInt32 PilotMain(UInt16 launchCode, MemPtr cmdPBP, UInt16 launchFlags) 
{
  Err err = 0;
  
  if (launchCode == sysAppLaunchCmdNormalLaunch) {
    if ((err = StartApplication()) == 0) {
      EventLoop();
      StopApplication();
    }
  }
  return err;
}

void InitMainForm(FormPtr pForm)
{
  ListPtr popupDays,popupToStations,popupFromStations;
  Int32 i=0,j=0;
  UInt16 nNumRecords, uCardNum;
  LocalID lID;
  DmOpenRef pdbfd, hAbbrevDb;
  MemHandle rec_mh, hAbbrevRec;
  char *recordP;
  DmSearchStateType searchState; 

  pdbfd = DmOpenDatabase (CardNo,pdbid,dmModeReadOnly);

  MemSet (StationNames, sizeof(StationNames), 0);
  numstations=0;

  /* Get number of records in database */
  NumRecords = DmNumRecords(pdbfd);

  /* populate StationNames from database */
  rec_mh = DmQueryRecord (pdbfd,STATIONRECORD);
  recordP = MemHandleLock (rec_mh);

  /* populate StationNames from database */
  i=0;
  j=0;
  while (recordP[i] != '\0') {
    if(recordP[i] == '\t') {
      StationNames[numstations][j]='\0';
      numstations++;
      j=0;
    }else {
      StationNames[numstations][j]=recordP[i];
      /* FIXME: What about some error handling ? */
      j++;
    }
    i++;
  }

  StationNames[numstations][j]='\0';
  numstations++;

  /* If 'DUMMY' is the last station this database is the 3 column format. */ 
  if ( StrNCompare( "DUMMY", StationNames[ numstations - 1 ], 5 ) == 0 )
  {
      numstations--;
      bUse3Cols = true;
      nIDofTable = resTimeTable3c;    /* Use 3 column table. */

      uCardNum = 0;

      //lID = DmFindDatabase( 0, "abbreviations" );
      if ( 0 == DmGetNextDatabaseByTypeCreator( true, &searchState, ABBR_PDB, CREATORID, false, &uCardNum, &lID ) )
      {
          hAbbrevDb = DmOpenDatabase( uCardNum, lID, dmModeReadOnly );
          if ( hAbbrevDb != 0 )
          {
              hAbbrevRec = DmQueryRecord( hAbbrevDb, 0 );
              StrCopy( szAbbreviations, MemHandleLock( hAbbrevRec ) );
              MemHandleUnlock( hAbbrevRec );
              /* Convert '|' to '\n'. */
              uCardNum = StrLen( szAbbreviations );
              for ( i = 0; i < uCardNum; i++ )
                  if ( szAbbreviations[ i ] == '|' )
                      szAbbreviations[ i ] = '\n';

              nNumRecords = DmNumRecords( hAbbrevDb );
              /* Read in the abbreviations. */
              for ( i = 1; i <= MAX_ABBREVIATIONS; i++ )
              {
                  /* Default to an empty string. */
                  StrCopy( Abbreviations[ i ], "" );

                  /* Ensure index is in range because some releases will */
                  /* display fatal error message if DmQueryRecords() called */
                  /* an out-of-range record number. */
                  if ( i <= nNumRecords )
                  {
                      hAbbrevRec = DmQueryRecord( hAbbrevDb, i );
                      if ( hAbbrevRec != NULL )
                      {
                          StrCopy( Abbreviations[ i ], MemHandleLock( hAbbrevRec ) );
                          MemHandleUnlock( hAbbrevRec );
                      }
                  }
              }
          }
      }
      else
      {
          StrCopy( szAbbreviations, "EWD = Mon to Sat\nM-F = Monday to Friday\nT-F = Tuesday to Friday\nMO/WO/FO/SO = Mon/Wed/Fri/Sat only\nWX/FX = Except Wed/Fri\nFSX = Fri/Sat excluded\nFSO = Fri/Sat only\nM-TH = Mon to Thurs" );

          /* Use defaults if database not found. */
          for ( i = 1; i <= MAX_ABBREVIATIONS; i++ )
              StrCopy( Abbreviations[ i ], "" );
          StrCopy( Abbreviations[ 1 ], "EWD" );
          StrCopy( Abbreviations[ 2 ], "M-F" );
          StrCopy( Abbreviations[ 3 ], "FO" );
          StrCopy( Abbreviations[ 4 ], "SO" );
          StrCopy( Abbreviations[ 5 ], "T-F" );
          StrCopy( Abbreviations[ 6 ], "MO" );
          StrCopy( Abbreviations[ 7 ], "WO" );
          StrCopy( Abbreviations[ 8 ], "WX" );
          StrCopy( Abbreviations[ 9 ], "FX" );
          StrCopy( Abbreviations[ 10 ], "FSX" );
          StrCopy( Abbreviations[ 11 ], "FSO" );
          StrCopy( Abbreviations[ 12 ], "M-TH" );
      } 
  }
  else
  {
      bUse3Cols = false;
      nIDofTable = resTimeTable;    /* Use 2 column table. */
  }

  for(i=0;i<numstations;i++)
    StationList[i]=StationNames[i];

  popupToStations=FrmGetObjectPtr(pForm,(FrmGetObjectIndex (pForm,resFromList)));
  popupFromStations=FrmGetObjectPtr(pForm,(FrmGetObjectIndex (pForm,resToList)));
  LstSetListChoices(popupToStations,StationList,numstations);
  LstSetListChoices(popupFromStations,StationList,numstations);

  /* set the selection to the value stored in preferences */
  if (FromStation != NOSELECTION) {
    LstSetSelection(popupFromStations,FromStation);
    CtlSetLabel(FrmGetObjectPtr(pForm,
                 FrmGetObjectIndex( pForm, resFromTrigger )),
                   LstGetSelectionText( popupFromStations, FromStation ));
  }

  if ( ToStation != NOSELECTION) {
    LstSetSelection(popupToStations,ToStation);
    CtlSetLabel(FrmGetObjectPtr( pForm,
                 FrmGetObjectIndex( pForm, resToTrigger )),
                  LstGetSelectionText( popupToStations, ToStation ));
  }

  /* release lock on record */
  MemHandleUnlock(rec_mh);

  /* populate DayList from database */
  rec_mh = DmQueryRecord (pdbfd,DAYRECORD);
  recordP = MemHandleLock (rec_mh);

  numdays=0;
  i=0;
  j=0;
  while (recordP[i] != '\0') {
    if(recordP[i] == '\t') {
      DayNames[numdays][j]='\0';
      numdays++;
      j=0;
    } else {
      DayNames[numdays][j]=recordP[i];
      j++;
    }
    i++;
  }

  DayNames[numdays][j]='\0';
  numdays++;

  for(i=0;i<numdays;i++)
    DayList[i]=DayNames[i];

  popupDays=FrmGetObjectPtr(gpForm,(FrmGetObjectIndex (pForm,resDayList)));
  LstSetListChoices(popupDays,DayList,numdays);

  /* set the selection to the value stored in preferences */
  if (Day != NOSELECTION) {
    LstSetSelection(popupDays,Day);
    CtlSetLabel(FrmGetObjectPtr( pForm,
                 FrmGetObjectIndex(pForm, resDayTrigger)),
                  LstGetSelectionText(popupDays, Day));
  }

  /* release lock on record */
  MemHandleUnlock(rec_mh);

  /* get offset for directions and dayx from database */
  rec_mh = DmQueryRecord (pdbfd,OFFSETRECORD);
  recordP = MemHandleLock (rec_mh);

  for(i=0; i<(numdays*2);i++)
    OffsetList[i] = (recordP[(i*2)+1] & 0xFF) | (recordP[i*2] << 8);

  /* release lock on record */
  MemHandleUnlock(rec_mh);

  DmCloseDatabase(pdbfd);

  FrmDrawForm(pForm);
}

/* initializes time table in Main Form */
void InitMainFormTable(FormPtr pForm)
{
  TablePtr tableP;
  Int16 numRows;
  RectangleType r;
  Int32 i=0,yoff;
  Coord rowheight;

  /* Hide both tables and show the required one. */
  /* Could do 'if ( nIDofTable == resTimeTable )' ... */
  FrmHideObject( gpForm, FrmGetObjectIndex( gpForm, resTimeTable ) );
  FrmHideObject( gpForm, FrmGetObjectIndex( gpForm, resTimeTable3c ) );
  FrmShowObject( gpForm, FrmGetObjectIndex( gpForm, nIDofTable ) );

  tableP = FrmGetObjectPtr(gpForm, 
                            FrmGetObjectIndex(gpForm,nIDofTable));
  numRows = TblGetNumberOfRows(tableP);


  /* Setup the rows of the table */
  for (i=0; i<numRows;i++) {
    TblSetItemStyle(tableP, i, FROMCOL, customTableItem);
    TblSetItemStyle(tableP, i, TOCOL, customTableItem); 
    if(bUse3Cols) TblSetItemStyle(tableP, i, INFOCOL, customTableItem);   /* daymobrew - Annotation column. */
    TblSetRowSelectable(tableP, i, false);
  }

  /* Make sure the columns are usable - so they draw */
  TblSetColumnUsable(tableP, FROMCOL, true);
  TblSetColumnUsable(tableP, TOCOL, true);
  if (bUse3Cols) TblSetColumnUsable(tableP, INFOCOL, true);    /* daymobrew - Annotation column. */
  TblHasScrollBar(tableP, true);

  /* Set our custom table item draw callback */
  TblSetCustomDrawProcedure(tableP, FROMCOL, TableDrawItemCallback);
  TblSetCustomDrawProcedure(tableP, TOCOL, TableDrawItemCallback);
  if (bUse3Cols) TblSetCustomDrawProcedure(tableP, INFOCOL, TableDrawItemCallback);    /* daymbrew - Annotation column. */

  /* Draw the outline of the table */
  TblGetBounds(tableP, &r);

  WinDrawGrayRectangleFrame(simpleFrame, &r);

  for (i = 1; i < numRows; i++) {
    rowheight=TblGetRowHeight(tableP,i);
    yoff = r.topLeft.y + 
             (i * rowheight) - 1;
    WinDrawGrayLine(r.topLeft.x, yoff,
                     r.extent.x,  yoff);
  }
}

/* Loads data from PDB file to Main Form Table */
void LoadMainFormTable(FormPtr pForm)
{
  DmOpenRef pdbfd;
  TablePtr tableP;
  MemHandle rec_mh;
  char *recordP;
  Int32 yoff,i,j;
  UInt16 numRows=0,time=0,offset=0,maxoffset=0,ScrollMax;
  Int8 n,ToS,FromS;
  RectangleType r;
  Coord rowheight;

  if((Day == NOSELECTION) || (FromStation == NOSELECTION)
     || (ToStation == NOSELECTION))
    return;

  if (FromStation > ToStation) { /* Southbound (Paris -> Banlieue) */
    offset = OffsetList[Day+numdays];
    ToS = numstations - ToStation - 1;
    FromS = numstations - FromStation - 1;
    if (Day != (numdays-1)) 
      maxoffset = OffsetList[Day+numdays+1] - 1;
    else
      maxoffset = NumRecords;
  } else {
    offset = OffsetList[Day];
    maxoffset = OffsetList[Day+1] - 1;

    FromS = FromStation;
    ToS = ToStation;
  }

  pdbfd = DmOpenDatabase (0,pdbid,dmModeReadOnly);

  tableP = FrmGetObjectPtr(pForm, 
                            FrmGetObjectIndex(pForm, nIDofTable));
  numRows = TblGetNumberOfRows(tableP);

  /* reset table data for security */
  i=0;
  while (i<numRows) {
    TblSetItemInt(tableP,i,FROMCOL, -1 );
    TblSetItemInt(tableP,i,TOCOL, -1 );
    if (bUse3Cols) TblSetItemInt(tableP,i,INFOCOL, -1 );    /* daymobrew - Annotation column. */
    i++;
  }

  /* populate the table with next data from PDB */
  i=0;
  j=0;
  ScrollMax = maxoffset-offset+1;
  while(i<numRows) {
    n=0;

   /* only get data if there are more datas */
    if ((ScrollValue+j+offset-1) < maxoffset) {
      rec_mh = DmQueryRecord (pdbfd, ScrollValue+j+offset-1);
      recordP = MemHandleLock (rec_mh);
    
      time = (recordP[(FromS*2)+1] & 0xFF) | (recordP[FromS*2] << 8);
      if(time!=65535) n++;
      TblSetItemInt(tableP,i,FROMCOL, time );
    
      time = (recordP[(ToS*2)+1] &0xFF) | (recordP[ToS*2] << 8);
      if(time!=65535) n++;
      TblSetItemInt(tableP,i,TOCOL, time );

      /* daymobrew - Annotation column. */
      if (bUse3Cols)
      {
          time = (recordP[(numstations*2)+1] &0xFF) | (recordP[numstations*2] << 8);
          TblSetItemInt(tableP,i,INFOCOL, time );
      }
      
      /* The train must stop at both station to be shown */
      if (n==2) {
        i++;
      } else {
        TblSetItemInt(tableP,i,FROMCOL, -1 );
        TblSetItemInt(tableP,i,TOCOL, -1 );
        if (bUse3Cols) TblSetItemInt(tableP,i,INFOCOL, -1 );    /* daymobrew - Annotation column. */
        ScrollMax--;
      }
      MemHandleUnlock(rec_mh);
    } else { /* no more data, the table must be filled */
      TblSetItemInt(tableP,i,FROMCOL, -1 );
      TblSetItemInt(tableP,i,TOCOL, -1 );
      if (bUse3Cols) TblSetItemInt(tableP,i,INFOCOL, -1 );    /* daymobrew - Annotation column. */
      i++;
    }
    
    j++;
  }

  TblDrawTable(tableP);
  
  /* Draw the outline of the rows */

  TblGetBounds(tableP, &r);

  for (i = 1; i < numRows; i++) {
    rowheight=TblGetRowHeight(tableP,i);
    yoff = r.topLeft.y + 
  //    (i * tableP->rowAttrs->height) - 1;
      (i * rowheight) - 1;
    WinDrawGrayLine(r.topLeft.x, yoff,
                     r.extent.x,  yoff);
  }


  if (ScrollValue > (ScrollMax-numRows))
    ScrollValue = ScrollMax-numRows;

  /* Update the scroll bar */
  SclSetScrollBar(FrmGetObjectPtr(pForm, 
                   FrmGetObjectIndex(pForm, 
                    resTableScrollBar )), 
                    ScrollValue,       /* current value */
                    0,                 /* min */
                    ScrollMax >= numRows ? ScrollMax - numRows : numRows,  /* max */
                    numRows );         /* page size */

  DmCloseDatabase(pdbfd);
}

/* Draws time schedule in cell */
void TableDrawItemCallback(void * table, Int16 row, Int16 column,
                             RectanglePtr bounds )
{
  TablePtr tableP = (TablePtr)table;
  UInt16 data;
  char str[6]={0};
  RectangleType rect;
  UInt8 Hour,Minute;
  char temp[8]={0};
  Int32 i;

  /* Erase previous label */
  rect = *bounds; 
  rect.topLeft.x += 1;
  rect.topLeft.y += 1;
  rect.extent.y  -= 1;
  WinEraseRectangle( &rect, 0 );

  data=TblGetItemInt(tableP,row,column);

    /* daymobrew - Handle Annotation column differently. Shouldn't be hard coded. */
    if ((bUse3Cols) && (column==INFOCOL))
    {
      if ( data && ( data < MAX_ABBREVIATIONS ) )
          StrCopy( str, Abbreviations[ data ] );
      else
          StrCopy( str, "" ); 
    }
    else
    {
      if(data==-1) {
        /* no data was found */
        StrPrintF(str,"    ");
      } else if (data==65535) { /* No train */
        StrPrintF(str," -- ");
      } else {

        StrIToA(str,data);

        if (StrLen(str) == 1) { /* train is between 00:00 and 00:09 */
          Hour = 0;
          Minute = str[1];
        } else if (StrLen(str) == 2) { /* train is between 00:10 and 00:59 */
          Hour = 0;
          Minute = StrAToI(str);
        } else {
          i = StrLen(str);
          temp[1] = str[i-1];
          temp[0] = str[i-2];
          temp[2] = '\0';
          Minute = (StrAToI(temp));
      
          str[i-2] = '\0';

          Hour = StrAToI(str);
        }

		/* Maybe I should use TimeToAscii for that */
    
        StrIToA(str, Hour);
        StrIToA(temp, Minute);
        StrCat(str,":");
        if(Minute < 10) StrCat(str, "0");
        StrCat(str, temp);
    }

  }

  /* Draw new label */
  FntSetFont(stdFont);
  WinDrawChars(str, StrLen(str), bounds->topLeft.x+2, bounds->topLeft.y);
}

/***********************************************************************
 *
 * FUNCTION:    MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:  command  - menu item id
 *
 * RETURNED:    true if the command was handled
 *
 * Original taken from the Puzzle example
 ************************************************************************/
static Boolean MainFormDoCommand (UInt16 command)
{
  Boolean handled = false;
  FormPtr pForm;

  MenuEraseStatus (0);

  switch (command) {
    case MainOptionsAboutCmd:
      pForm = FrmInitForm(aboutForm);
      FrmDoDialog(pForm);
      handled = true;
      break;

    case MainOptionsSelectCmd:
      pForm = FrmInitForm(selectForm);

      FrmSetEventHandler(pForm, (FormEventHandlerPtr)selectFormEventHandler);
      FrmGotoForm(selectForm);
      handled = true;
      break;

    case MainOptionsAbbreviations:
      FrmCustomAlert( alertAbbreviations, szAbbreviations, "", "" );
      handled = true;
      break;

    default:
      break;
  }
  return handled;
}

/* Function SetFormLabel
 *
 * Sets the label of the form with current time
 *
 */
void SetFormLabel(FormPtr pForm)
{
  UInt32 seconds=0;
  DateTimePtr dateTimeP;
  char newtitle[32]={0};
  UInt16 len=0;
  char *asctime;

  if (pForm==NULL) return;

  seconds = TimGetSeconds();

  dateTimeP = MemPtrNew(sizeof(DateTimeType));

  TimSecondsToDateTime(seconds,dateTimeP);

  len = StrLen(DBName);

  if(len>31) len = 31; /* bad hack */

  StrNCopy(newtitle,DBName,len);

  StrCat(newtitle,"  ");

  asctime=MemPtrNew(10);
  TimeToAscii(dateTimeP->hour,dateTimeP->minute,tfColon24h,asctime);

  StrCat(newtitle,asctime);

  FrmSetTitle(pForm,newtitle);

  MemPtrFree(asctime);
  MemPtrFree(dateTimeP);
}

/* Draws items in selection table */
void SelTableDrawItem(void * table, Int16 row, Int16 column,
                       RectanglePtr bounds )
{
  TablePtr tableP = (TablePtr)table;
  char str[dmDBNameLength]={0};
  RectangleType rect;
  dbrowType *data;
  
  /* Erase previous label */
  rect = *bounds; 
  rect.topLeft.x += 1;
  rect.topLeft.y += 1;
  rect.extent.y  -= 1;
  WinEraseRectangle( &rect, 0 );
 
  data = TblGetItemPtr(tableP,row,column);
  
  if (data)
    StrCopy (str, data->DBName);
  else
    TblSetRowSelectable( tableP, row, false );

  /* Draw new label */
  FntSetFont(stdFont);
  WinDrawChars(str, StrLen(str), bounds->topLeft.x+2, bounds->topLeft.y);
}

/* Initialize The selectForm Table */
void InitSelTable(FormPtr pForm)
{
  RectangleType r;
  TablePtr tableP;
  Int32 i;
  UInt16 numRows=0;

  tableP = FrmGetObjectPtr(pForm, 
                            FrmGetObjectIndex(pForm,selectFrmTable));
  numRows = TblGetNumberOfRows(tableP);

  /* Setup the rows of the table */
  for(i=0; i<numRows;i++) {
    TblSetItemStyle( tableP, i, SELECTCOL, customTableItem );
    TblSetRowSelectable( tableP, i, true );
  }

  /* Make sure the columns are usable - so they draw */
  TblSetColumnUsable( tableP, SELECTCOL, true );
  TblHasScrollBar( tableP, true );

  /* Set our custom table item draw callback */
  TblSetCustomDrawProcedure( tableP, SELECTCOL, SelTableDrawItem );

  /* Draw the outline of the table */
  TblGetBounds(tableP, &r);

  WinDrawRectangleFrame( simpleFrame, &r );
}

/* Populates and draw selectForm table 
 * Update : false if this is the first call, true for a refresh (scrolling...)
*/
void LoadSelTable(FormPtr pForm, Boolean Update)
{
  TablePtr tableP;
  Int16 i=0, j=0, oldscrollvalue=0;
  UInt16 numRows=0;
  Boolean newSearch = true;
  DmSearchStateType searchState;
  UInt16 CardNo;
  LocalID dbID;
  char name[dmDBNameLength];
  dbrowType *pdbarrayP;

	tableP = FrmGetObjectPtr(pForm, 
															FrmGetObjectIndex(pForm,selectFrmTable));
	numRows = TblGetNumberOfRows(tableP);

	oldscrollvalue = ScrollSelectValue;

	if (Update == false) {
		/* let's count how many databases we have in memory */
		while (0 == DmGetNextDatabaseByTypeCreator(newSearch, &searchState,
                                               PDBTYPE, CREATORID,
                                               false, &CardNo, &dbID)) {
			selpdbcount++;
			newSearch=false;
		}

		if (selpdbcount > 0) {
			pdbarrayH = MemHandleNew(selpdbcount * sizeof(dbrowType));
			pdbarrayP=MemHandleLock(pdbarrayH);

			newSearch = true;
			selpdbcount=0;
			while (0 == DmGetNextDatabaseByTypeCreator(newSearch, &searchState,
																								 PDBTYPE, CREATORID,
																								 false, &CardNo, &dbID)) {
				newSearch=false;

				DmDatabaseInfo(CardNo, dbID, name, NULL, NULL, NULL, NULL, NULL,
												NULL, NULL, NULL, NULL, NULL);

				MemMove(pdbarrayP[selpdbcount].DBName, name, dmDBNameLength);
				pdbarrayP[selpdbcount].CardNo = CardNo;
				pdbarrayP[selpdbcount].pdbid = dbID;

				selpdbcount++;
			}
		/*	
			while (i < selpdbcount) {
				TblSetItemPtr(tableP,i,SELECTCOL,(dbrowType*)pdbarrayP++);
				i++;
			}
		*/
		}
	} else {
			pdbarrayP=MemHandleLock(pdbarrayH);
	}

	/* avoid scrolling too much */
	if ((selpdbcount - ScrollSelectValue) < numRows) {
		if (selpdbcount > numRows) {
			ScrollSelectValue = selpdbcount-numRows;
		} else {
			ScrollSelectValue = 0;
		}
	}

	
	i=j=0;
	while ((i < selpdbcount) && (j < numRows)) {
		if ((i < ScrollSelectValue) || (i > (ScrollSelectValue+numRows))) {
			  pdbarrayP++;
		} else {
				TblSetItemPtr(tableP,j,SELECTCOL,(dbrowType*)pdbarrayP++);
				TblSetRowUsable(tableP, j, true);
				j++;
		}
		i++;
	}

	if (j < numRows) {
		while (j < numRows) {
			TblSetRowUsable(tableP, j, false);
			j++;
		}
	}

	SclSetScrollBar(FrmGetObjectPtr( pForm, 
									 FrmGetObjectIndex( pForm, 
										selectFrmTableScrollBar )), 
										oldscrollvalue,          /* current value */
										//0,          /* min */
                    selpdbcount > numRows ? 0 : selpdbcount,  /* min */
										selpdbcount,    /*max*/
										numRows );  /* page size */

	TblEraseTable(tableP);
	TblDrawTable(tableP);
}

/* Callback to ExgDBWrite for beaming */
Err WriteDBData(const void *dataP, UInt32 *sizeP, void *userDataP)
{
  Err err;
  
  *sizeP = ExgSend((ExgSocketPtr)userDataP, (void*)dataP, *sizeP, &err);
  return err;
}
